//
//  AmityChatViewController.swift
//  SampleApp
//
//  Created by Federico Zanetello on 8/7/19.
//  Copyright © 2019 David Zhang. All rights reserved.
//

import UIKit
import Photos
import MobileCoreServices
import AmitySDK

private let reuseIdentifier = "SampleAppTableViewCell"

final class AmityChatViewController: UIViewController,
                                     UIImagePickerControllerDelegate,
                                     UINavigationControllerDelegate,
                                     UITextViewDelegate,
                                     UIDocumentPickerDelegate,
                                     KeyboardServiceDelegate,
                                     CommentsDelegate,
                                     AmityCustomViewControllerDelegate {
    
    private lazy var titleButton: UIButton = {
        let button = UIButton(type: .system)
        button.addTarget(self,
                         action: #selector(updateChannelDisplayName),
                         for: .touchUpInside)
        button.clipsToBounds = true
        return button
    }()
    
    @IBOutlet private weak var textView: UITextView!
    @IBOutlet private weak var inputContainer: UIView!
    @IBOutlet private weak var bottom: NSLayoutConstraint!
    @IBOutlet private weak var membersTableView: UITableView!
    @IBOutlet private weak var tableViewHeightConstraint: NSLayoutConstraint!
    @IBOutlet weak var sendButton: UIButton!
    
    private var messagesTableViewController: AmityMessagesTableViewController?
    
    //private var channelType: AmityChannelType = AmityChannelType.standard
    
    private var mentionManager: AmityMentionManager?
    
    // To be injected.
    var client: AmityClient!
    
    // To be injected.
    var channelId: String = ""
    var subChannel: AmitySubChannel?
    var channelType: AmityChannelType = .community
    
    var msgObject: AmityObject<AmityMessage>?
    var msgToken: AmityNotificationToken?
    
    lazy var messageRepository = AmityMessageRepository(client: self.client)
    lazy var channelRepository = AmityChannelRepository(client: self.client)
    lazy var fileRepository = AmityFileRepository(client: self.client)
    lazy var channelParticipation = AmityChannelParticipation(client: client,
                                                         andChannel: channelId)
    
    var messageObject: AmityObject<AmityMessage>?
    var messageToken: AmityNotificationToken?
    
    let separatorView = UIView()
    
    // MARK: Lifecycle
    
    override func viewWillAppear(_ animated: Bool) {
        super.viewWillAppear(animated)
        KeyboardService.shared.delegate = self
        Task {
            do {
                try await subChannel?.startReading()
            }
        }
        mentionManager?.delegate = self
    }
    
    override func viewWillDisappear(_ animated: Bool) {
        super.viewWillDisappear(animated)
        Task {
            do {
                try await subChannel?.stopReading()
            }
        }
        mentionManager?.delegate = nil
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        textView.delegate = self
        setPlaceholder()
        
        navigationItem.titleView = titleButton
        addLeaveButton()
        addSettingsButton()
        
        let nib = UINib(nibName: "SampleAppTableViewCell", bundle: nil)
        membersTableView?.register(nib, forCellReuseIdentifier: reuseIdentifier)
        
        membersTableView?.tableFooterView = UIView()
        hideMentionsTableView()
        
        setupViews()
        
        setupMentionManager()
        
        textView.autocorrectionType = .no
    }
    
    private func setupViews() {
        separatorView.translatesAutoresizingMaskIntoConstraints = false
        inputContainer.addSubview(separatorView)
        NSLayoutConstraint.activate([
            separatorView.heightAnchor.constraint(equalToConstant: 1),
            separatorView.leadingAnchor.constraint(equalTo: inputContainer.leadingAnchor),
            separatorView.topAnchor.constraint(equalTo: inputContainer.topAnchor),
            separatorView.trailingAnchor.constraint(equalTo: inputContainer.trailingAnchor)
        ])
        separatorView.backgroundColor = UIColor.systemGroupedBackground
        
        sendButton.setImage(UIImage(systemName: "paperplane.fill"), for: .normal)
        sendButton.setTitle("", for: .normal)
        
        textView.addRoundedCorner(radius: 4)
        textView.backgroundColor = UIColor.systemGroupedBackground
        textView.textContainerInset = UIEdgeInsets(top: 11, left: 8, bottom: 0, right: 8)
    }
    
    private func setupMentionManager() {
        mentionManager = AmityMentionManager(withType: .message(channelId: channelId))
        mentionManager?.delegate = self
    }
    
    override func prepare(for segue: UIStoryboardSegue, sender: Any?) {
        super.prepare(for: segue, sender: sender)
        switch segue.destination {
        case let messagesTableViewController as AmityMessagesTableViewController:
            self.messagesTableViewController = messagesTableViewController
            messagesTableViewController.client = client
            messagesTableViewController.channelId = channelId
            messagesTableViewController.subChannelId = subChannel?.id ?? ""
        default:
            break
        }
    }
    
    private func hideMentionsTableView() {
        tableViewHeightConstraint.constant = 0
        membersTableView?.isHidden = true
    }
    
    @IBAction private func showImagePickerRequest(_ sender: UIButton) {
        checkAuthorizationForPicker(isVideo: false)
    }
    
    @IBAction private func showVideoPickerRequest(_ sender: UIButton) {
        checkAuthorizationForPicker(isVideo: true)
    }
    
    private func checkAuthorizationForPicker(isVideo: Bool) {
        switch PHPhotoLibrary.authorizationStatus() {
        case .notDetermined:
            PHPhotoLibrary.requestAuthorization { [weak self] _ in
                self?.checkAuthorizationForPicker(isVideo: isVideo)
            }
        case .denied,
             .restricted:
            Log.add(info: "No authorization 🙈")
        case .authorized:
            if isVideo {
                showVideoPicker()
            } else {
                showImagePicker()
            }
        @unknown default:
            break
        }
    }
    
    private func showVideoPicker() {
        let alertController = UIAlertController()
        
        if UIImagePickerController.isSourceTypeAvailable(.camera) {
            let cameraVideoAction = UIAlertAction(title: "Camera", style: .default) { [weak self]_ in
                let videoPicker = UIImagePickerController()
                videoPicker.sourceType = .camera
                videoPicker.mediaTypes = [String(kUTTypeMovie)]
                videoPicker.videoExportPreset = AVAssetExportPresetPassthrough
                videoPicker.delegate = self
                self?.present(videoPicker, animated: true, completion: nil)
            }
            alertController.addAction(cameraVideoAction)
        }
        if UIImagePickerController.isSourceTypeAvailable(.photoLibrary) {
            let videoPickerAction = UIAlertAction(title: "Gallery", style: .default) { [weak self]_ in
                let videoPicker = UIImagePickerController()
                videoPicker.sourceType = .photoLibrary
                videoPicker.mediaTypes = [String(kUTTypeMovie)]
                videoPicker.videoExportPreset = AVAssetExportPresetPassthrough
                videoPicker.delegate = self
                self?.present(videoPicker, animated: true, completion: nil)
            }
            alertController.addAction(videoPickerAction)
        }
        alertController.addAction(UIAlertAction(title: "Cancel", style: .cancel, handler: nil))
        DispatchQueue.main.async {
            self.present(alertController, animated: true)
        }
    }
    
    private func showImagePicker() {
        guard UIImagePickerController.isSourceTypeAvailable(.photoLibrary) else { return }
        DispatchQueue.main.async {
            
            let imagePicker = UIImagePickerController()
            imagePicker.sourceType = .photoLibrary
            imagePicker.delegate = self
            
            self.present(imagePicker, animated: true, completion: nil)
        }
    }
    
    @IBAction private func didPressSend(_ sender: UIButton) {
        sendTextMessage()
    }
    
    // MARK: UIImagePickerControllerDelegate
    func imagePickerController(_ picker: UIImagePickerController, didFinishPickingMediaWithInfo info: [UIImagePickerController.InfoKey: Any]) {
        picker.dismiss(animated: true, completion: nil)
                
        if let imageUrl = info[.imageURL] as? URL {
            sendImageMessage(imageUrl)
            
        } else if let videoUrl = info[.mediaURL] as? URL {
            sendVideoMessage(videoUrl)
        }
    }
    
    private func sendImageMessage(_ imageURL: URL) {
        
        let options = AmityImageMessageCreateOptions(subChannelId: subChannel?.id ?? "", imageURL: imageURL, fullImage: true, parentId: messagesTableViewController?.replyToMessageId)
        Task {
            do {
                let _ = try await messageRepository.createImageMessage(options: options)
            } catch {
                self.displayError(title: "Upload Image error", error: error)
            }
        }
        
        messagesTableViewController?.replyToMessageId = nil
    }
    
    private func sendVideoMessage(_ videoURL: URL) {
        let options = AmityVideoMessageCreateOptions(subChannelId: subChannel?.id ?? "", videoFileURL: videoURL, parentId: messagesTableViewController?.replyToMessageId)
        Task {
            do {
                let _ = try await messageRepository.createVideoMessage(options: options)
            } catch {
                self.displayError(title: "Upload Video error", error: error)
            }
        }
        messagesTableViewController?.replyToMessageId = nil
    }
    
    private func sendTextMessage() {
        guard let text: String = textView.text, !text.isEmpty else { return }
        textView.text = ""

        var parentId: String?
        if let messagesTableViewController = self.messagesTableViewController {
            parentId = messagesTableViewController.replyToMessageId
            messagesTableViewController.replyToMessageId = nil
        }
        
        messageToken?.invalidate()
        let metadata = mentionManager?.getMetadata()
        let mentionees = mentionManager?.getMentionees()
        
        let options = AmityTextMessageCreateOptions(subChannelId: subChannel?.id ?? "", text: text, tags: nil, parentId: parentId, metadata: metadata, mentioneesBuilder: mentionees)
        Task {
            do {
               let _ = try await messageRepository.createTextMessage(options: options)
            } catch {
                self.displayError(title: "Message Send Error", error: error)
            }
        }
          
        mentionManager?.resetState()
        messageToken = messageObject?.observe({ (msg, error) in
            Log.add(info: "Data Status: \(msg.dataStatus.rawValue)")
            Log.add(info: "isEdited: \(String(describing: msg.object?.isEdited))")
            
        })
    }
    
    func joinChannel(channelId: String, type: AmityChannelType) {
        // Don't join to channel if type is conversation.
        if type == .conversation {
            return
        }
        
        Task {
            do {
                let channel = try await channelRepository.joinChannel(channelId: subChannel?.id ?? "")
                setDisplayName(channel.displayName)
            } catch let error as NSError where error.code == 400304  {
                presentUserBanError()
            }
        }
    }
    
    private func displayError(title: String, error: Error) {
        let alertController = UIAlertController(
            title: title,
            message: error.localizedDescription,
            preferredStyle: .alert
        )
        let defaultAction = UIAlertAction(title: "OK", style: .cancel) { _ in
            self.dismiss(animated: true, completion: nil)
        }
        alertController.addAction(defaultAction)
        present(alertController, animated: true, completion: nil)
    }
    
    @MainActor
    private func presentUserBanError() {
        let alertController = UIAlertController(
            title: "Can't join",
            message: "User is banned from a channel",
            preferredStyle: .alert
        )
        let defaultAction = UIAlertAction(title: "OK", style: .cancel) { _ in
            self.dismiss(animated: true, completion: nil)
            self.navigationController?.popViewController(animated: true)
        }
        alertController.addAction(defaultAction)
        present(alertController, animated: true, completion: nil)
    }
    
    func seeComments(for parentId: String) {
        UserDefaults.standard.parentId = parentId
        guard
            let viewController: UIViewController = storyboard?.instantiateViewController(withIdentifier: "AmityComments"),
            let chatViewController = viewController as? AmityChatViewController
        else { return }
        chatViewController.channelId = channelId
        chatViewController.subChannel = subChannel
        chatViewController.client = client
        chatViewController.joinChannel(channelId: channelId, type: .standard)
        navigationController?.pushViewController(chatViewController, animated: true)
    }
    
    private func addLeaveButton() {
        let leaveButton = UIBarButtonItem(title: "Leave",
                                          style: .plain,
                                          target: self,
                                          action: #selector(leave))
        navigationItem.setLeftBarButton(leaveButton,
                                        animated: true)
        navigationItem.leftItemsSupplementBackButton = true
    }
    
    private func addSettingsButton() {
        let gearGlyph = UIImage(named: "Settings")
        
        let settingsButton = UIBarButtonItem(image: gearGlyph,
                                             style: .plain,
                                             target: self,
                                             action: #selector(navigateToChannelSettingsView))
        let removeAvatarButton = UIBarButtonItem(image: UIImage(systemName: "xmark"), style: .plain, target: self, action: #selector(removeChannelAvatar))
        let addAvatarButton = UIBarButtonItem(image: UIImage(systemName: "person"), style: .plain, target: self, action: #selector(addChannelAvatar))
        navigationItem.setRightBarButtonItems([settingsButton, removeAvatarButton, addAvatarButton], animated: true)
    }
    
    @objc private func navigateToChannelSettingsView() {
        let storyboard = UIStoryboard(name: "ChannelSettings", bundle: nil)
        let viewController = storyboard
            .instantiateViewController(withIdentifier: "ChannelSettingsTableViewController")
        guard
            let channelSettingsViewController = viewController as? ChannelSettingsTableViewController
        else { return }
        
        channelSettingsViewController.client = client
        channelSettingsViewController.subChannelId = subChannel?.id ?? ""
        channelSettingsViewController.channelId = channelId
        channelSettingsViewController.subChannel = subChannel
        
        navigationController?.pushViewController(channelSettingsViewController,
                                                 animated: true)
    }
    
    @objc private func leave() {
        Task {
            do {
                try await channelRepository.leaveChannel(channelId: channelId)
                navigationController?.popViewController(animated: true)
            } catch {
                // Even though showLeaveError is marked with @MainActor, it still does not run on main thread
                // unless explicitly called on main thread
                await MainActor.run {
                    showLeaveError(error)
                }
            }
        }
    }
    
    @MainActor
    private func showLeaveError(_ error: Error) {
        var message: String {
            if error.isAmityErrorCode(.moderatorUnableToLeaveCommunity) {
                return "You’re the only moderator in this group. To leave community, nominate other members to moderator role."
            } else {
                return error.localizedDescription
            }
        }
        
        AppUtility.showAlert(in: self, title: "Unable to leave community", message: message)
    }
    
    // add title button action
    @objc private func updateChannelDisplayName() {
        let alertController = UIAlertController(title: "Enter The New Channel Display Name",
                                                message: nil,
                                                preferredStyle: .alert)
        
        let renameAction = UIAlertAction(title: "Rename", style: .default) { [weak self] _ in
            
            guard
                let textField: UITextField = alertController.textFields?.first,
                let textFieldText: String = textField.text,
                let channelId = self?.channelId else { return }
            
            Task { [weak self] in
                do {
                    let builder = AmityChannelUpdateBuilder(channelId: channelId)
                    builder.setDisplayName(textFieldText)
                    let _ = try await self?.channelRepository.editChannel(with: builder)
                    self?.setDisplayName(textFieldText)
                } catch {
                    print("update channel displayname error: \(error.localizedDescription)")
                }
            }
            
        }
        
        alertController.addAction(renameAction)
        
        alertController.addTextField { [weak self] textField in
            textField.placeholder = self?.titleButton.title(for: .normal)
        }
        
        let cancelAction = UIAlertAction(title: "Cancel",
                                         style: .cancel,
                                         handler: nil)
        alertController.addAction(cancelAction)
        
        present(alertController, animated: true, completion: nil)
    }
    
    // add remove avatar button action
    @objc private func removeChannelAvatar() {
        let alertController = UIAlertController(title: "Remove Channel Avatar",
                                                message: nil,
                                                preferredStyle: .alert)
        
        let removeAction = UIAlertAction(title: "Remove image", style: .default) { [weak self] _ in
            guard let strongSelf = self else {
                return
            }
            Task {
                do {
                    let builder = AmityChannelUpdateBuilder(channelId: strongSelf.channelId)
                    builder.setAvatar(nil)
                    let _ = try await strongSelf.channelRepository.editChannel(with: builder)
                    Log.add(info: "Update Status: success")
                } catch {
                    Log.add(info: "Update Status: \(error.localizedDescription)")
                }
            }
        }
        let cancelAction = UIAlertAction(title: "Cancel", style: .cancel, handler: nil)
        
        alertController.addAction(removeAction)
        alertController.addAction(cancelAction)
        
        present(alertController, animated: true, completion: nil)
    }
    
    @objc private func addChannelAvatar() {
        let alertController = UIAlertController(title: "Add Channel Avatar",
                                                message: nil,
                                                preferredStyle: .alert)
        
        let removeAction = UIAlertAction(title: "Add clock image", style: .default) { [weak self] _ in
            guard let self else { return }
            Task { @MainActor in
                do {
                    let result = try await self.fileRepository.uploadImage(UIImage(systemName: "clock")!, progress: nil)
                    let builder = AmityChannelUpdateBuilder(channelId: self.channelId)
                    builder.setAvatar(result)
                    let _ = try await self.channelRepository.editChannel(with: builder)
                    Log.add(info: "Avatar Update: success")
                } catch {
                    Log.add(info: "Avatar Update Error: \(error.localizedDescription)")
                }
            }
        }
        
        let cancelAction = UIAlertAction(title: "Cancel", style: .cancel, handler: nil)
        
        alertController.addAction(removeAction)
        alertController.addAction(cancelAction)
        
        present(alertController, animated: true, completion: nil)
    }
    
    @MainActor
    private func setDisplayName(_ channelDisplayName: String?) {
        let newTitleText: String
        if
            let channelName: String = channelDisplayName,
            !channelName.isEmpty {
            newTitleText = channelName
        } else {
            newTitleText = "@\(channelId)"
        }
        titleButton.setTitle(newTitleText, for: .normal)
        titleButton.sizeToFit()
    }
    
    private func setPlaceholder() {
        textView.text = "Enter message here"
        textView.textColor = UIColor.lightGray
    }
    
    func textViewDidBeginEditing(_ textView: UITextView) {
        if textView.textColor == UIColor.lightGray {
            textView.text = nil
            textView.textColor = UIColor.black
        }
    }
    
    func textViewDidEndEditing(_ textView: UITextView) {
        if textView.text.isEmpty {
            setPlaceholder()
        }
    }
    
    func textView(_ textView: UITextView, shouldChangeTextIn range: NSRange, replacementText text: String) -> Bool {
        if (textView.text ?? "").count > AmityMentionManager.maximumCharacterCountForPost {
            showAlertForMaximumCharacters()
            return false
        }
        
        return mentionManager?.shouldChangeTextIn(textView, inRange: range, replacementText: text, currentText: textView.text ?? "") ?? true
    }
    
    func textViewDidChangeSelection(_ textView: UITextView) {
        mentionManager?.changeSelection(textView)
    }
    
    // MARK: KeyboardServiceDelegate
    
    func keyboardWillChange(service: KeyboardService,
                            height: CGFloat,
                            animationDuration duration: TimeInterval) {
        let offset = height > 0 ? view.safeAreaInsets.bottom : 0
        bottom.constant = -height + offset
        
        view.setNeedsUpdateConstraints()
        view.layoutIfNeeded()
        
        for child in children {
            switch child {
            case let messagesTableViewController as AmityMessagesTableViewController:
                messagesTableViewController.scrollToBottom()
            default:
                break
            }
        }
    }
    
    @IBAction func handleFileUploadButton(_ sender: Any) {
        let importMenu = UIDocumentPickerViewController(documentTypes: [String(kUTTypePDF)], in: .import)
        importMenu.delegate = self
        importMenu.modalPresentationStyle = .fullScreen
        present(importMenu, animated: true, completion: nil)
    }
    
    @IBAction func handleCustomButton(_ sender: Any) {
        let vc = AmityCustomViewController.makeViewController()
        (vc as? AmityCustomViewController)?.delegate = self
        navigationController?.pushViewController(vc, animated: true)
    }
    
    public func documentPicker(_ controller: UIDocumentPickerViewController, didPickDocumentsAt urls: [URL]) {
        guard let myURL = urls.first else { return }
        
        let options = AmityFileMessageCreateOptions(subChannelId: subChannel?.id ?? "", fileURL: myURL, parentId: messagesTableViewController?.replyToMessageId)
        
        Task {
            do {
                let _ = try await messageRepository.createFileMessage(options: options)
            } catch {
                self.displayError(title: "Upload File error", error: error)
            }
        }
        
        messagesTableViewController?.replyToMessageId = nil
    }
    
    func documentPickerWasCancelled(_ controller: UIDocumentPickerViewController) {
        dismiss(animated: true, completion: nil)
    }
    
    // MARK: - AmityCustomViewControllerDelegate
    
    func amityCustom(_ viewController: AmityCustomViewController, willSendCustomDataWithData data: [String : Any]) {
        let options = AmityCustomMessageCreateOptions(subChannelId: subChannel?.id ?? "", data: data, parentId: messagesTableViewController?.replyToMessageId)
        Task {
            do {
                let _ = try await messageRepository.createCustomMessage(options: options)
            } catch {
                self.displayError(title: "Message Create error", error: error)
            }
        }
        
    }
    
    func amityCustom(_ viewController: AmityCustomViewController, willSendVoiceMessageWithData audioFileURL: URL, fileName: String) {
        let options = AmityAudioMessageCreateOptions(subChannelId: subChannel?.id ?? "", audioFileURL: audioFileURL, parentId: messagesTableViewController?.replyToMessageId)
        Task {
            do {
                let _ = try await messageRepository.createAudioMessage(options: options)
            } catch {
                self.displayError(title: "Upload Audio error", error: error)
            }
        }
        
        messagesTableViewController?.replyToMessageId = nil
    }
    
    func amityCustom(_ viewController: AmityCustomViewController, willUpdateCustomDataWithData data: [String : Any], onMessage message: AmityMessage?) {
        // Intentionally left empty
    }
}

extension AmityChatViewController: UITableViewDelegate {
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        mentionManager?.addMention(from: textView, in: textView.text ?? "", at: indexPath)
    }
}

extension AmityChatViewController: UITableViewDataSource {
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        return mentionManager?.itemsCount ?? 0
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        guard let sampleAppCell = tableView.dequeueReusableCell(withIdentifier: reuseIdentifier) as? SampleAppTableViewCell else { return UITableViewCell() }
        if let model = mentionManager?.item(at: indexPath) {
            sampleAppCell.titleLabel?.text = model.isChannel ? "@all: Notify everyone in this channel" : model.displayName
        }
        
        return sampleAppCell
    }
    
    func tableView(_ tableView: UITableView, heightForRowAt indexPath: IndexPath) -> CGFloat {
        return 40.0
    }
    
    func tableView(_ tableView: UITableView, willDisplay cell: UITableViewCell, forRowAt indexPath: IndexPath) {
        if indexPath.row == (mentionManager?.itemsCount ?? 0) - 4 {
            mentionManager?.loadMore()
        }
    }
}

extension AmityChatViewController: AmityMentionManagerDelegate {
    func didGetUsers(users: [AmityMentionUserModel]) {
        if users.isEmpty {
            tableViewHeightConstraint.constant = 0
            membersTableView.isHidden = true
        } else {
            var heightConstant:CGFloat = 240.0
            if users.count < 5 {
                heightConstant = CGFloat(users.count) * 52.0
            }
            tableViewHeightConstraint.constant = heightConstant
            membersTableView.isHidden = false
            membersTableView.reloadData()
        }
    }
    
    func didCreateAttributedString(attributedString: NSAttributedString) {
        textView.attributedText = attributedString
    }
    
    func didMentionsReachToMaximumLimit() {
        let message = "Mentions are reached to maximum limit"
        AppUtility.showAlert(in: self, title: "", message: message)
    }
    
    func didCharactersReachToMaximumLimit() {
        showAlertForMaximumCharacters()
    }
    
    private func showAlertForMaximumCharacters() {
        let title = "Unable to mention"
        let message = "Text reached to maximum characters limit"
        AppUtility.showAlert(in: self, title: title, message: message)
    }
}
