//
//  MediaListController.swift
//  SampleApp
//
//  Created by Nutchaphon Rewik on 28/3/2565 BE.
//  Copyright © 2565 BE David Zhang. All rights reserved.
//

import AVKit
import UIKit
import AmitySDK
import MobileCoreServices

protocol MediaListControllerDelegate: AnyObject {
    func mediaListControllerDidSaveMediaList(_ controller: MediaListController)
}

extension MediaListController {
    
    enum MediaType {
        
        case video
        case image
        case file
        
        var title: String {
            switch self {
            case .video:
                return "Videos"
            case .image:
                return "Images"
            case .file:
                return "Files"
            }
        }
        
    }
    
    
}

class MediaListController: UIViewController {

    weak var delegate: MediaListControllerDelegate?
    
    var videos: [AmityVideoData] = []
    var images: [AmityImageData] = []
    var files: [AmityFileData] = []
    
    let mediaType: MediaType
    
    private var fileRepository: AmityFileRepository?
    
    @IBOutlet private weak var tableView: UITableView!
    private var saveButtonItem: UIBarButtonItem!
    private var filePicker: FilePicker?
    private var progressView: UIProgressView?
    
    private var dataChanges: Bool = false {
        didSet {
            updateSaveButtonState()
        }
    }
    
    init(mediaType: MediaType) {
        self.mediaType = mediaType
        fileRepository = AmityFileRepository(client: AmityManager.shared.client!)
        super.init(nibName: "MediaListController", bundle: .main)
    }
    
    required init?(coder: NSCoder) {
        fatalError("init(coder:) has not been implemented")
    }
    
    override func viewDidLoad() {
        super.viewDidLoad()
        title = mediaType.title
        saveButtonItem = UIBarButtonItem(barButtonSystemItem: .save, target: self, action: #selector(saveChanges))
        navigationItem.rightBarButtonItems = [
            UIBarButtonItem(barButtonSystemItem: .add, target: self, action: #selector(addMedia)),
            saveButtonItem
        ]
        tableView.register(UITableViewCell.self, forCellReuseIdentifier: "cell")
        tableView.dataSource = self
        tableView.delegate = self
        updateSaveButtonState()
    }
    
    @objc private func addMedia() {
        switch mediaType {
        case .video:
            showVideoPicker()
        case .image:
            showImagePicker()
        case .file:
            showFilePicker()
        }
    }
    
    private func updateSaveButtonState() {
        saveButtonItem.isEnabled = dataChanges
    }
    
    @objc private func saveChanges() {
        delegate?.mediaListControllerDidSaveMediaList(self)
        navigationController?.popViewController(animated: true)
    }
        
    private func showVideoPicker() {
        let filePickerAction = UIAlertAction(title: "File Manager", style: .default) { _ in
            self.filePicker = FilePicker(presentationController: self, delegate: self)
            self.filePicker?.maximumSizeOfIFiles = 4 * 1_073_741_824 // 4GB
            self.filePicker?.present(from: self.view, documentTypes: ["public.data"])
        }
        let videoPickerAction = UIAlertAction(title: "Photo Album", style: .default) { _ in
            guard UIImagePickerController.isSourceTypeAvailable(.photoLibrary) else { return }
            let videoPicker = UIImagePickerController()
            videoPicker.sourceType = .photoLibrary
            videoPicker.mediaTypes = [String(kUTTypeMovie)]
            videoPicker.videoExportPreset = AVAssetExportPresetPassthrough
            videoPicker.delegate = self
            self.present(videoPicker, animated: true, completion: nil)
        }
        
        let alertController = UIAlertController()
        alertController.addAction(filePickerAction)
        alertController.addAction(videoPickerAction)
        alertController.addAction(UIAlertAction(title: "Cancel", style: .destructive))
        present(alertController, animated: true)
        
    }
    
    private func showImagePicker() {
        guard UIImagePickerController.isSourceTypeAvailable(.photoLibrary) else { return }
        let imagePicker = UIImagePickerController()
        imagePicker.sourceType = .photoLibrary
        imagePicker.mediaTypes = [String(kUTTypeImage)]
        imagePicker.delegate = self
        present(imagePicker, animated: true, completion: nil)
    }
    
    private func showFilePicker() {
        filePicker = FilePicker(presentationController: self, delegate: self)
        filePicker?.present(from: view)
    }
    
    private func uploadVideo(from localURL: URL) {
        let alertViewController = UIAlertController(title: "Uploading", message: nil, preferredStyle: .alert)
        present(alertViewController, animated: true) {
            //  Add your progressbar after alert is shown (and measured)
            let margin:CGFloat = 8.0
            let rect = CGRect(x: margin, y: 72.0, width: alertViewController.view.frame.width - margin * 2.0 , height: 2.0)
            self.progressView = UIProgressView(frame: rect)
            self.progressView?.progress = 0.5
            self.progressView?.tintColor = self.view.tintColor
            alertViewController.view.addSubview(self.progressView!)
        }
        Task { @MainActor in
            do {
                let video = try await fileRepository?.uploadVideo(with: localURL, progress: { [weak self] progress in
                    self?.progressView?.progress = Float(progress)
                })
                if let video = video {
                    self.videos.append(video)
                    self.dataChanges = true
                    self.dataSourceDidUpdate()
                    alertViewController.dismiss(animated: true, completion: nil)
                }
            } catch {
                print("upload error: \(error.localizedDescription)")
                alertViewController.dismiss(animated: true) {
                    self.showErrorAlert(message: error.localizedDescription)
                }
            }
        }
    }
    
    private func showErrorAlert(message: String) {
        let alertController = UIAlertController(title: nil, message: message, preferredStyle: .alert)
        let okAction = UIAlertAction(title: "OK", style: .default)
        alertController.addAction(okAction)
        present(alertController, animated: true)
    }
    
    private func uploadImage(from image: UIImage) {
        let alertController = UIAlertController(title: "Uploading", message: nil, preferredStyle: .alert)
        present(alertController, animated: true, completion: nil)
        Task { @MainActor in
            do {
                let result = try await fileRepository?.uploadImage(image, progress: { progress in
                    Log.add(info: "Progress: \(progress)")
                })
                alertController.dismiss(animated: true, completion: nil)
                if let result {
                    Log.add(info: "Image upload id: \(result.fileId)")
                    self.images.append(result)
                    self.dataChanges = true
                    self.dataSourceDidUpdate()
                }
            } catch {
                alertController.dismiss(animated: true, completion: nil)
                print("upload error: \(error.localizedDescription)")
            }
        }
    }
    
    private func uploadFile(from localURL: URL) {
        let alertController = UIAlertController(title: "Uploading", message: nil, preferredStyle: .alert)
        present(alertController, animated: true, completion: nil)
        let fileName = localURL.lastPathComponent
        
        let resources = try? localURL.resourceValues(forKeys:[.fileSizeKey, .typeIdentifierKey])
        let fileSize = resources?.fileSize ?? 0
        let typeIdentifier = resources?.typeIdentifier ?? ""
        
        print("-> file url \(localURL.path)")
        print("-> file size \(fileSize) | identifier \(typeIdentifier)")
        Task { @MainActor [weak self] in
            do {
                let fileData = try await fileRepository?.uploadFile(with: localURL, fileName: fileName, progress: nil)
                alertController.dismiss(animated: true, completion: nil)
                guard let fileData else { return }
                
                self?.files.append(fileData)
                self?.dataChanges = true
                self?.dataSourceDidUpdate()
                print("-> upload success")
            } catch {
                alertController.dismiss(animated: true, completion: nil)
                print("-> upload error: \(error.localizedDescription)")
            }
        }
        
    }
    
    private func promptActionsAt(index: Int) {
        
        let alertController = UIAlertController(title: "Choose the action", message: nil, preferredStyle: .actionSheet)
        let delete = UIAlertAction(title: "Delete", style: .destructive) { [weak self] action in
            guard let strongSelf = self else { return }
            switch strongSelf.mediaType {
            case .video:
                strongSelf.videos.remove(at: index)
            case .image:
                strongSelf.images.remove(at: index)
            case .file:
                strongSelf.files.remove(at: index)
            }
            self?.dataChanges = true
            self?.dataSourceDidUpdate()
        }
        let cancel = UIAlertAction(title: "Cancel", style: .cancel, handler: nil)
        alertController.addAction(delete)
        alertController.addAction(cancel)
        
        present(alertController, animated: true, completion: nil)
    }
    
    private func dataSourceDidUpdate() {
        tableView.reloadData()
    }
    
}

extension MediaListController: UITableViewDataSource {
    
    func numberOfSections(in tableView: UITableView) -> Int {
        return 1
    }
    
    func tableView(_ tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        switch mediaType {
        case .video:
            return videos.count
        case .image:
            return images.count
        case .file:
            return files.count
        }
    }
    
    func tableView(_ tableView: UITableView, cellForRowAt indexPath: IndexPath) -> UITableViewCell {
        let cell = tableView.dequeueReusableCell(withIdentifier: "cell", for: indexPath)
        var content = cell.defaultContentConfiguration()
        switch mediaType {
        case .video:
            let video = videos[indexPath.row]
            content.text = video.fileId
        case .image:
            let image = images[indexPath.row]
            content.text = image.fileId
        case .file:
            let file = files[indexPath.row]
            content.text = file.fileId
        }
        cell.contentConfiguration = content
        return cell
    }
    
}

extension MediaListController: UITableViewDelegate {
    
    func tableView(_ tableView: UITableView, didSelectRowAt indexPath: IndexPath) {
        tableView.deselectRow(at: indexPath, animated: true)
        promptActionsAt(index: indexPath.row)
    }
    
}

extension MediaListController: UINavigationControllerDelegate, UIImagePickerControllerDelegate {
    
    func imagePickerController(_ picker: UIImagePickerController, didFinishPickingMediaWithInfo info: [UIImagePickerController.InfoKey: Any]) {
        
        if let mediaType = info[.mediaType] as? String {
            switch mediaType {
            case String(kUTTypeImage):
                if let image: UIImage = info[UIImagePickerController.InfoKey.originalImage] as? UIImage {
                    DispatchQueue.main.async { [weak self] in
                        picker.dismiss(animated: true) {
                            self?.uploadImage(from: image)
                        }
                    }
                }
            case String(kUTTypeMovie):
                if let videoUrl = info[.mediaURL] as? URL {
                    DispatchQueue.main.async { [weak self] in
                        picker.dismiss(animated: true) {
                            self?.uploadVideo(from: videoUrl)
                        }
                    }
                }
            default:
                break
            }
        }
        
    }
    
}

extension MediaListController: FilePickerDelegate {
    func didPickFiles(filesUrl: [URL], documentTypes: [String]) {
        guard let fileUrl = filesUrl.first else { return }
        if documentTypes.contains("public.data") {
            uploadVideo(from: fileUrl)
        } else {
            uploadFile(from: fileUrl)
        }
    }
}
