//
//  PostLiveObjectPage.swift
//  SampleApp
//
//  Created by Nutchaphon Rewik on 23/9/2565 BE.
//  Copyright © 2565 BE David Zhang. All rights reserved.
//

import SwiftUI
import AmitySDK

struct PostLiveObjectPage: View {
    
    private enum AlertType {
        case postIdCopied
        case hardDeleting
        case editingText
        case actionError(Error)
    }
    
    // MARK: -
    let postId: String
    
    // MARK: - View States
    @State private var liveObject: AmityObject<AmityPost>?
    
    // This state indicates that the user already unobserve the curernt live object.
    @State private var unobserved = false
    
    @State private var snapshot: AmityPost?
    @State private var isLoading = false
    @State private var error: Error?
    
    @State private var token: AmityNotificationToken?
    
    private var loadingColor: Color {
        (isLoading) ? .yellow : .clear
    }
    
    private var errorColor: Color {
        (error != nil) ? .red : .clear
    }
    
    @State private var hasStartObserved = false
    
    // MARK: - Alert States
    @State private var displayingAlert: AlertType? {
        didSet {
            if displayingAlert != nil {
                if (showingAlert) {
                    showingAlert = false
                }
                showingAlert = true
            } else {
                showingAlert = false
            }
        }
    }
    @State var showingAlert: Bool = false
    
    var body: some View {
        VStack(alignment: .leading, spacing: 8) {
            Text("postId: \(postId)")
                .underline()
                .padding(.top, 10)
                .onTapGesture {
                    UIPasteboard.general.string = postId
                    displayingAlert = .postIdCopied
                }
            if !hasStartObserved {
                Text("observe(👁) live object to see the data")
                    .foregroundColor(Color.gray)
                    .font(.caption)
            } else {
                HStack {
                    Button("Unobserve") {
                        unobserve()
                    }
                    .compatibleBorderedButtonStyle()
                    .disabled(unobserved)
                }
                if unobserved {
                    Text(
"""
You've unobserved the current live object.
To see data updates, please create the new live object and observe with (👁) button.
"""
                    )
                    .foregroundColor(Color.gray)
                    .font(.caption)
                }
                Text("isLoading: \(isLoading ? "true" : "false")")
                    .background(loadingColor)
                Text("error: \(error?.localizedDescriptionWithCode ?? "<nil>")")
                    .background(errorColor)
                VStack (alignment: .leading) {
                    Text("Snapshot")
                        .font(.body)
                        .padding(.leading, 3.0)
                    VStack(alignment: .leading) {
                        if let snapshot = snapshot {
                            Text("postId: \(snapshot.postId)")
                                .font(.caption)
                            Text("dataType: \(snapshot.dataType)")
                                .font(.caption)
                            Text("text: \(snapshot.data?["text"] as? String ?? "<nil>")")
                                .font(.caption)
                            Text("isDeleted: \(snapshot.isDeleted ? "true": "false")")
                                .font(.caption)
                            Text("created: \(DateFormatter.localizedString(from: snapshot.createdAt, dateStyle: .short, timeStyle: .medium))")
                                .font(.caption)
                            Text("updated: \(DateFormatter.localizedString(from: snapshot.updatedAt, dateStyle: .short, timeStyle: .medium))")
                                .font(.caption)
                        } else {
                            Text("<nil>")
                                .font(.caption)
                        }
                    }
                    .padding(.leading, 8)
                }
                .border(.blue, width: 1.0)
            }
            Spacer()
        }
        .padding([.leading, .trailing], 8.0)
        .frame(maxWidth: .infinity, alignment: .topLeading)
        .navigationTitle("Live Object")
        .toolbar {
            ToolbarItemGroup(placement: .navigationBarTrailing) {
                Button { hardDelete() } label: { Image(systemName: "trash") }
                Button { editText() } label: { Image(systemName: "pencil") }
                Button { observe() } label: { Image(systemName: "eye") }
            }
        }
        .alert(isPresented: $showingAlert) {
            switch displayingAlert! {
            case .postIdCopied:
                return Alert(title: Text("post id copied."))
            case .editingText:
                return Alert(title: Text("Editing using random text..."))
            case .hardDeleting:
                return Alert(title: Text("Hard deleting..."))
            case .actionError(let error):
                return Alert(
                    title: Text("An error occured"),
                    message: Text(error.localizedDescriptionWithCode)
                )
            }
        }
    }
    
    private func observe() {
        if !hasStartObserved {
            hasStartObserved = true
        }
        liveObject = AmityManager.shared.postRepository?.getPost(withId: postId)
        unobserved = false
        token = liveObject?.observe { liveObject, error in
            updateStates()
        }
    }
    
    private func hardDelete() {
        Task { @MainActor in
            displayingAlert = .hardDeleting
            do {
                let _ = try await AmityManager.shared.postRepository?.hardDeletePost(withId: postId, parentId: nil)
                displayingAlert = nil
            } catch {
                displayingAlert = .actionError(error)
            }
        }
    }
    
    private func editText() {
        // builder
        let builder = AmityTextPostBuilder()
        let randomText = "random-text: \(UUID().uuidString.prefix(8))"
        builder.setText(randomText)
        //
        displayingAlert = .editingText
        Task { @MainActor in
            do {
                let result = try await AmityManager.shared.postRepository?.editPost(withId: postId, builder: builder)
                displayingAlert = nil
            } catch {
                displayingAlert = .actionError(error)
            }
        }
    }
    
    private func updateStates() {
        guard let liveObject = liveObject else {
            assertionFailure("liveObject must have a value")
            return
        }
        snapshot = liveObject.object
        error = liveObject.error
        switch liveObject.loadingStatus {
        case .loaded:
            isLoading = false
        default:
            isLoading = true
        }
    }
    
    private func unobserve() {
        token?.invalidate()
        unobserved = true
    }
    
}

struct PostLiveObject_Previews: PreviewProvider {
    static var previews: some View {
        PostLiveObjectPage(postId: "<test>")
    }
}

private extension Error {
    var localizedDescriptionWithCode: String {
        switch self {
        case let nsError as NSError:
            return "(\(nsError.code)) \(nsError.localizedDescription)"
        default:
            return localizedDescription
        }
    }
}
